import os
import json
from pydantic import BaseModel, Field
from mq_http_sdk.mq_client import MQClient
from mq_http_sdk.mq_producer import TopicMessage

from awe.util.logging import logger
from awe.types.common import Payload, ResponseMessage
from awe.util.util import debounce
from awe.util.workflow import convert_to_old_data, convert_to_old_result

class MQProducerOptions(BaseModel):
  endpoint: str = Field(title="MQ Endpoint", description="MQ Endpoint", default="")
  access_key: str = Field(title="MQ Access Key", description="MQ Access Key", default="")
  access_secret: str = Field(title="MQ Access Secret", description="MQ Access Secret", default="")
  instance_id: str = Field(title="实例ID", description="实例ID", default="")
  is_static_topic: bool = Field(title="是否为静态Topic", description="是否为静态Topic", default=False)
  env: str = Field(title="环境", description="环境", default="")
  appname: str = Field(title="应用名", description="应用名", default="")
  response_topic: str = Field(title="响应Topic", description="响应Topic", default="")

class MQProducer:
  def __init__(self, options: MQProducerOptions = None):
    self.options = options or MQProducerOptions(
      endpoint=os.getenv('ROCKETMQ_ENDPOINT'),
      access_key=os.getenv('ROCKETMQ_ACCESS_KEY'),
      access_secret=os.getenv('ROCKETMQ_ACCESS_SECRET'),
      instance_id=os.getenv('ROCKETMQ_INSTANCE_ID'),
      is_static_topic=os.getenv('ROCKETMQ_STATIC_TOPIC'),
      env=os.getenv('ENV'),
      appname=os.getenv('APPNAME'),
      response_topic=os.getenv('ROCKETMQ_BASE_RESPONSE_TOPIC')
    )
    self.producer = self.get_producer(self.options)
    
  def get_producer(self, options: MQProducerOptions):
    mq_client = MQClient(
      options.endpoint,
      options.access_key,
      options.access_secret
    )
    if options.is_static_topic:
      producer_topic = options.response_topic
    else:
      topic_prefix = f"{options.env.upper()}_{options.appname.upper()}"
      producer_topic = f"{topic_prefix}_{options.response_topic}"
    
    logger.info(f'producer_topic is {producer_topic}')
    return mq_client.get_producer(options.instance_id, producer_topic)

  def send_progress_message(self, message: ResponseMessage, payload: Payload):
    message_body = {
      "code": "0",
      "status": "PROCESSING",
      "success": True,
      "taskId": payload.taskId,
      "userId": payload.userId,
      "clientId": payload.clientId,
      "client_id": payload.clientId,
      "appName": self.options.appname,
      "response": message.response,
      "workflowCode": payload.workflowCode
    }
    # 兼容旧版本，response_message 中要包含 progress
    if payload.messageTag != 'WEB':
      message_body["progress"] = message.response["progress"]
    debounced = message.get("debounced")
    if debounced:
      self.debounce_progress_message(json.dumps(message_body), payload.messageTag)
    else:
      self.publish_message(json.dumps(message_body), tag=payload.messageTag)
    
  @debounce(0.1)
  def debounce_progress_message(self, message: str, tag: str):
    self.publish_message(message, tag)

  def send_error_message(self, message: ResponseMessage, payload: Payload):
    message_body = {
      "code": message.code if message.code else "500",
      "status": "FAILED",
      "success": False,
      "taskId": payload.taskId,
      "userId": payload.userId,
      "clientId": payload.clientId,
      "client_id": payload.clientId,
      "appName": self.options.appname,
      "response": message.response,
      "workflowCode": payload.workflowCode
    }
    self.publish_message(json.dumps(message_body), tag=payload.messageTag)

  def send_complete_message(self, message: ResponseMessage, payload: Payload):
    code = f"{message.code}" if message.code else "0"
    status = "SUCCEED" if code == "0" else "FAILED"
    message_body = {
      "code": code,
      "status": status,
      "success": status == "SUCCEED",
      "taskId": payload.taskId,
      "userId": payload.userId,
      "clientId": payload.clientId,
      "client_id": payload.clientId,
      "appName": self.options.appname,
      "workflowCode": payload.workflowCode,
      "response": message.response,
    }
    if payload.messageTag == 'WEB':
      message_body['results'] = []
    # 兼容老版本，results 是 response 的 data 的 images 的 output
    try:
      message_body['results'] = convert_to_old_result(message)
    except Exception as e:
      message_body['results'] = []
      logger.error(f'[MQProducer] for compatability, get image output from response error: {e}')
    # 兼容老版本，message_body 中包含 data
    try:
      message_body['data'] = convert_to_old_data(message)
    except Exception as e:
      message_body['data'] = {}
      logger.error(f'[MQProducer] for compatability, get image output from response error: {e}')
    self.publish_message(json.dumps(message_body), tag=payload.messageTag)
  
  def publish_message(self, message: str, tag: str = None):
    try:
      topic_message = TopicMessage(message, tag)
      re_msg = self.producer.publish_message(topic_message)
      logger.info("Publish Succeed, TopicName:%s MessageID:%s MessageBody:%s" % (self.options.response_topic, re_msg.message_id, message))
    except Exception as e:
      logger.error("Publish, Message Fail. Exception:%s" % e)



import json
import requests
from awe.util.logging import logger

def post_request(server_address, endpoint, data=None):
    url = f"http://{server_address}{endpoint}"
    headers = {"Content-Type": "application/json"} if data else {}
    json_data = json.dumps(data).encode("utf-8") if data else None
    response = requests.post(url, headers=headers, data=json_data)
    if response.status_code != 200:
        logger.info(f"Failed: {endpoint}, status code: {response.status_code}")
    return response


import os
import re
import sys
import random
import base64
import traceback
from io import BytesIO

from awe.oss import get_download_url, get_file_content, upload_file
from awe.types.common import Payload, ResponseMessage
from awe.util.util import generate_filename_with_date, get_file_extension, get_remote_base64_image, get_remote_json, upload_image_to_comfyui, upload_remote_image
from awe.util.logging import logger
from awe.util.tongyi import translate as tongyi_translate
from awe.util.comfyui_exception import ComfyUIException

def contains_chinese(check_str):
  return re.search('[\u4e00-\u9fff]', check_str)

def translate(text):
  logger.info(f"translate text: {text}")
  if contains_chinese(text):
    logger.info(f"contains chinese, use tongyi translate: {text}")
    return tongyi_translate(text)
  else:
    logger.info(f"not contains chinese, use origin text: {text}")
  return text

def get_remote_workflow_from_oss(workflow_code):
  env = os.getenv('ENV')
  appname = os.getenv('APPNAME')
  file_path = f"ai-workflow-engine/workflows/{env}/{appname}/{workflow_code}.json"
  return get_file_content(file_path)

def get_workflow_data(payload: Payload):
  task_id = payload.get('taskId')
  # 这里的 workflow 并非 ComfyUI 的 workflow，而是包含 paramTpl, outputTpl 和 contentTpl 的 workflow
  workflow_data = payload.get('workflow')
  # 如果 workflow_data 为空，尝试从 workflowUrl 获取
  if workflow_data is None:
    workflow_url = payload.get('workflowUrl')
    # 如果 workflow_url 不为空，尝试从远程获取
    if workflow_url is not None:
      workflow_data = get_remote_json(workflow_url)
    # 如果从 workflowUrl 获取失败，尝试从 workflowCode 获取
    if workflow_data is None:
      workflow_code = payload.get('workflowCode')
      if workflow_code is not None:
        workflow_data = get_remote_workflow_from_oss(workflow_code)
  # 如果 workflow_data 为空，返回错误
  if workflow_data is None:
    logger.error(f"get workflow data failed, task_id: {task_id}")
    return None
  logger.info(f"workflow_data: {workflow_data}")
  return workflow_data

def build_workflow(workflow_data, params=None):
  if params is None:
    params = workflow_data.get('params')
  
  logger.info(f"workflow_data: {workflow_data}")
  logger.info(f"params: {params}")
  
  content_tpl = workflow_data.get('contentTpl')
  param_tpl = workflow_data.get('paramTpl')
  params = params or workflow_data.get('params')
  logger.info(f"workflow params: {params}")
  if params is None:
    logger.error("workflow params is None, just return content_tpl")
    return content_tpl
  if param_tpl is None:
    logger.error("workflow paramTpl is None, just return content_tpl")
    return content_tpl
  
  for key in param_tpl:
    param_node = param_tpl.get(f'{key}')
    node_id = param_node.get("nodeId")
    param_key = param_node.get("paramKey")
    param_type = param_node.get("paramType")
    param = params.get(f'{key}')
    node = content_tpl.get(f'{node_id}')
    if node is None:
      logger.error(f"node {node_id} is None, skip it.")
      continue
    inputs = node.get("inputs")
    if param_type is not None and param_type.upper() == "RANDOM":
      if inputs[f'{param_key}'] is None:
        inputs[f'{param_key}'] = random.randint(0, 1125899906842624)
      continue
      
    if param is None:
      logger.warning(f"param {key} is None")
      continue
    node = build_workflow_node(key, param_node, params, node, node_id)
    content_tpl[f'{node_id}'] = node
  
  return content_tpl

def build_workflow_node(key, param_node, params, node, node_id):
  inputs = node.get("inputs")
  param = params.get(f'{key}')
  param_key, param_type, enable_translate = (param_node.get("paramKey"), 
    param_node.get("paramType"), param_node.get("enableTranslate"))
  if param_type is None:
    logger.error(f"[WorkflowUtil] paramType is None, skip it, node_id: {node_id}, key: {key}")
    return node
  logger.info(f"[WorkflowUtil] param_type: {param_type}")
  if param_type.upper() == "STRING" or param_type.upper() == "TEXT":
    if enable_translate == 'Y':
      inputs[f'{param_key}'] = translate(param)
    else:
      inputs[f'{param_key}'] = param
  elif param_type.upper() == "IMAGE":
    build_image_node(param, inputs, node, node_id)
  elif param_type.upper() in ["INT", "FLOAT", "NUMBER"]:
    inputs[f'{param_key}'] = param
  elif param_type.upper() == "ARRAY":
    build_array_node(key, param_key, param_node, param, inputs)
  else:
    logger.warning(f"[WorkflowUtil] paramType not support: {param_type}")

  node["inputs"] = inputs
  return node

def build_image_node(param, inputs, node, node_id):
  handled = node.get('handled')
  if handled:
    logger.info(f"[WorkflowUtil] node {node_id} has been handled")
    return
  host = os.getenv('COMFYUI_HOST')
  port = os.getenv('COMFYUI_PORT')
  img_url = param
  class_type = node['class_type']
  if class_type == 'Base64ImageInput':
    inputs['base64_image'] = get_remote_base64_image(img_url)
  else:
    logger.info(f"[WorkflowUtil] image url: {img_url}")
    if img_url is None or type(img_url) is not str:
      logger.error(f"image is not url, node_id: {node_id}, type: {node['class_type']}")
      return
    elif img_url.startswith("http"):
      new_image_url = img_url
    else:
      new_image_url = get_download_url(img_url)
      logger.info(f"new_image_url : {new_image_url}")
    upload_url = f'http://{host}:{port}/upload/image'
    resp = upload_image_to_comfyui(upload_url, 'image.png', new_image_url)
    if resp is not None:
      inputs['image'] = resp.get('name')
      node['handled'] = True
  
def build_array_node(key, param_key, param_node, param, inputs):
  param_sub_type = param_node.get("paramSubType")
  if type(param) is not list:
    if (param_sub_type is None and type(param) is str):
      inputs[f'{param_key}'] = param
      return
    logger.error(f"param {key} is not list, but paramSubType is Array, skip it.")
    return
  if param_sub_type == "Image":
    inputs[f'{param_key}'] = [get_remote_base64_image(p) for p in param]
  else:
    inputs[f'{param_key}'] = param
         
def limit_input_value(input_key, inputs, max_value, options=None):
  if options is None:
    options = {}
  node = options.get('node')
  node_id = options.get('node_id')
  
  if isinstance(inputs[input_key], str):
    inputs[input_key] = int(inputs[input_key])
  
  if isinstance(inputs[input_key], (int, float)):
    if inputs[input_key] > max_value:
      inputs[input_key] = max_value
      logger.warning(f"Limiting {input_key} to {max_value}, {inputs}")
      raise ComfyUIException(f"{input_key} 不能大于 {max_value}", node_id, node.get('class_type', 'Unknown'))
 
def _handle_image_input(inputs, new_inputs, node, node_id):
  try:
    build_image_node(inputs['image'], new_inputs, node, node_id)
  except Exception as e:
    logger.error(f"[WorkflowUtil] handle image failed: {e}")
    raise ComfyUIException("图片有误，请重新上传", node_id, node.get('class_type', 'Unknown'))

def _process_input_key(key, inputs, new_inputs, node, node_id):
  if key == "image":
    _handle_image_input(inputs, new_inputs, node, node_id)
  elif key in ["seed", "noise_seed", "rand_seed"]:
    control_after_generate = inputs.get('control_after_generate') or 'randomize'
    if control_after_generate == 'fixed':
      new_inputs[key] = inputs[key]
    else:
      new_inputs[key] = random.randint(0, 1125899906842624)
  elif key == 'steps':
    limit_input_value(key, new_inputs, 40, {'node_id': node_id, 'node': node})
  elif key in ['scale', 'scale_by', 'upscale_by']:
    limit_input_value(key, new_inputs, 2, {'node_id': node_id, 'node': node})
  elif key in ['batch_size', 'num_samples', 'num_samples_per_class', 'num_classes', 'num_classes_per_batch', 'num_classes_per_sample']:
    limit_input_value(key, new_inputs, 4, {'node_id': node_id, 'node': node})
  elif key in ['height', 'empty_latent_height', 'width', 'empty_latent_width', 'side_length', 'image_gen_width', 'image_gen_height']:
    limit_input_value(key, new_inputs, 2048, {'node_id': node_id, 'node': node})
  else:
    logger.warning(f"key {key} not support")

def handle_workflow_inputs(workflow):
  for node_id, node in workflow.items():
    try:
      inputs = node.get("inputs", {})
      new_inputs = {**inputs}
      
      for key in inputs:
        _process_input_key(key, inputs, new_inputs, node, node_id)
      
      node["inputs"] = new_inputs
      workflow[node_id] = node
        
    except Exception as e:
      logger.error(f"[WorkflowUtil] handle_workflow_inputs failed: {e}")
      logger.error(f"[WorkflowUtil] handle_workflow_inputs failed: {traceback.print_exc()}")
      if isinstance(e, ComfyUIException):
        raise e
      raise ComfyUIException(f"handle_workflow_inputs failed: {e}", node_id, node.get('class_type', 'Unknown'))
  
  return workflow
  
def build_remote_image_url():
  host = os.getenv('COMFYUI_HOST')
  port = os.getenv('COMFYUI_PORT')
  return f"http://{host}:{port}/view"
      
def _handle_dict_item(item, file_extension, compress=True):
  if item.get("filename"):
    return _handle_filename_item(item, file_extension, compress)
  elif item.get("id"):
    return _handle_id_item(item, compress)
  else:
    logger.error(f"filename not found in item: {item}")
  return item

def _handle_filename_item(item, file_extension, compress=True):
  if item.get('url') is not None and item.get('url').startswith("http"):
    logger.info(f"item url is {item.get('url')}, return it")
    return item
  path = generate_filename_with_date(get_file_extension(item["filename"]))
  preview_url = build_remote_image_url()
  new_path, file_url = upload_remote_image(preview_url, item, path, file_extension, compress)
  return {**item, "url": file_url, "path": new_path}

def _handle_id_item(item, compress=True):
    result = {**item}
    urls = {
      "image_url": "png",
      "video_url": "mp4",
      "audio_url": "mp3"
    }
    
    for url_key, ext in urls.items():
      url = item.get(url_key)
      if url:
        path = generate_filename_with_date(ext)
        new_url, _ = upload_remote_image(url, item, path, ext, compress)
        result[url_key] = new_url
        if url_key == "image_url":
          result["path"] = new_url
    
    return result

def _handle_str_item(base64_image):
  if base64_image.startswith("http"):
    return {"url": base64_image, "path": base64_image}
  
  if base64_image.startswith("data:image"):
    path = generate_filename_with_date("png")
    image_bytes = base64.b64decode(base64_image)
    input_stream = BytesIO(image_bytes)
    file_url = upload_file(input_stream, path)
    return {"url": file_url, "path": path}
  
  logger.error(f"base64_image {base64_image} not support")
  return base64_image

def append_image_path(items, file_extension="png", compress=True):
  for i in range(len(items)):
    if isinstance(items[i], dict):
      items[i] = _handle_dict_item(items[i], file_extension, compress)
    elif isinstance(items[i], str):
      items[i] = _handle_str_item(items[i])
    else:
      logger.error(f"item {items[i]} not support")

def handle_image_result(comfyui_result, compress=True):
  try:
    for node_id in comfyui_result:
      node = comfyui_result.get(node_id)
      print(f"node: {node}")
      for key in node:
        item = node.get(key)
        if type(item) is dict:
          item = [item]
          append_image_path(item, compress)
          node[key] = item[0]
        elif type(item) is list:
          append_image_path(item, compress)
        else:
          logger.error(f"item {item} not support")
  except Exception as e:
    logger.error(f"[WorkflowUtil] handle_image_result failed: {e}")
    logger.error(f"[WorkflowUtil] handle_image_result failed: {traceback.print_exc()}")

def format_workflow_resp(resp, output_tpl):
  result_obj = {}
  for key in output_tpl:
    output_config = output_tpl.get(key)
    node_id = output_config.get("nodeId")
    param_key = output_config.get("paramKey")
    node = resp.get(f'{node_id}')
    result = None
    try:
      result = node.get(param_key)
    except Exception as e:
      logger.error(f"get result failed: {e}")
    try:
      if key == "images" or output_config.get("paramType") == "Image":
        if result is None and param_key != "gifs":
          result = node.get("gifs")
    except Exception as e:
      logger.error(f"get result failed: {e}")
    result_obj[key] = {
      "output": result,
      "title": output_config.get("name"),
      "paramType": output_config.get("paramType"),
    }
  return result_obj

def build_result(comfyui_result, workflow_data):
  # 根据 output_tpl 格式化结果
  output_tpl = workflow_data.get('outputTpl')
  result = format_workflow_resp(comfyui_result, output_tpl)
  return result

def convert_to_old_data(message: ResponseMessage):
  """
  兼容老版本，将 message 转换为老版本的 data 数据
  """
  data = {}
  try:
    message_data = message.response['data']
    if message_data:
      for key, value in message_data.items():
        data[key] = value['output']
  except Exception as e:
    logger.error(f'[MQProducer] for compatability, get image output from response error: {e}')
    data = {}
  return data

def convert_to_old_result(message: ResponseMessage):
  """
  兼容老版本，将 message 转换为老版本的 results 数据
  """
  results = []
  try:
    results = message.response['data']['images']['output'] or []
  except Exception as e:
    logger.error(f'[MQProducer] for compatability, get image output from response error: {e}')
    results = []
  return results



import json
import os
import piexif
import base64
import requests
from PIL import Image
from io import BytesIO
from threading import Timer
from nanoid import generate
from datetime import datetime
from awe.util.logging import logger
from awe.oss import upload_file

def get_remote_json(url):
  return requests.get(url).json()

def get_remote_base64_image(url):
  if url is None:
    logger.error("URL is None, cannot get image.")
    return None
  if url.startswith("http"):
    # 向URL发送GET请求
    response = requests.get(url)
    # 确保请求成功
    if response.status_code == 200:
      # 从响应中获取图片数据作为二进制
      image_data = response.content
      # 创建一个BytesIO对象
      image_bytes = BytesIO(image_data)
      # 使用Pillow打开图片
      image = Image.open(image_bytes)
      # 创建一个新的BytesIO对象，这次用于保存base64编码的数据
      buff = BytesIO()
      # 保存图片到buff对象中（格式可以使用image.format来自动确定）
      image.save(buff, format=image.format)
      # 获取buff内的二进制数据
      byte_data = buff.getvalue()
      # 进行base64编码并解码为UTF-8字符串
      base64_str = base64.b64encode(byte_data).decode('utf-8')
      # 返回base64编码的图片数据
      return base64_str
    else:
      raise requests.exceptions.HTTPError(f"Failed to get image from {url}, Status Code: {response.status_code}")
  
  # 判断是否是 base64 编码的图片
  try:
    # 尝试解码base64字符串
    base64.b64decode(url)
    # 如果没有抛出异常，则是base64编码的图片
    return url
  except Exception as e:
    logger.error(f"URL is not valid, cannot get image. {e}")
    return None

def generate_filename_with_date(file_format):
  file_id = generate('1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', 10)
  dir_path = datetime.now().strftime("%Y/%m/%d")
  return f"{dir_path}/{file_id}.{file_format}"

def get_file_extension(file_name):
  return file_name.split('.')[-1]

def create_dir_if_not_exists(directory):
  try:
    os.makedirs(directory, exist_ok=True)
    logger.info(f"确保目录 '{directory}' 存在。")
  except OSError as error:
    logger.error(f"创建目录时发生错误：{error}")
  return directory

def is_animated_webp(image):
  return getattr(image, 'is_animated', False)

def compress_png(image_path, output_path):
  try:
    # 打开图像文件
    with Image.open(image_path) as img:
      # 检查图像模式，如果是RGBA则转换为支持透明度的WebP
      if img.mode == 'RGBA':
        try:
          img.save(output_path, format='WEBP', lossless=True, transparency=img.info.get('transparency'))
        except Exception as e:
          img.save(output_path, format='WEBP')
      else:
        img.save(output_path, format='WEBP')
    print(f"转换成功，文件已保存至：{output_path}")
  except Exception as e:
    print(f"转换过程中发生错误：{e}")

def add_img_source(source_path, target_path, meta: dict):
  now = datetime.now().strftime("%Y:%m:%d %H:%M:%S")
  img_description = meta.get("description", "")
  img_artist = meta.get("artist", "")
  img_software = meta.get("software", "")
  img_copyright = meta.get("copyright", "")
  img_datetime = meta.get("DateTime", now)
  img_datetime_original = meta.get("DateTimeOriginal", now)
  comment = f"{img_description}\n{img_artist}\n{img_software}\n{img_copyright}"
  
  exif_info = {
    "0th": {
      piexif.ImageIFD.ImageDescription: comment,
      piexif.ImageIFD.Artist: img_artist,
      piexif.ImageIFD.Software: img_software,
      piexif.ImageIFD.Copyright: img_copyright,
      piexif.ImageIFD.DateTime: img_datetime,
    },
    "Exif": {
      piexif.ExifIFD.DateTimeOriginal: img_datetime_original,
    }
  }

  # 试图把EXIF信息转为字节流
  try:
      exif_bytes = piexif.dump(exif_info)
      if exif_bytes is not None:
        # 读取图片
        img = Image.open(source_path)
        # 把EXIF信息写入图片
        img.save(target_path, exif=exif_bytes)
  except ValueError as e:
      print(f"遇到错误：{e}")
      exif_bytes = b""

def upload_remote_image(url, params, path, file_extension='png', compress=True):
  IMAGE_META_COPYRIGHT = os.getenv("IMAGE_META_COPYRIGHT")
  IMAGE_META_ARTIST = os.getenv("IMAGE_META_ARTIST")
  IMAGE_META_SOFTWARE = os.getenv("IMAGE_META_SOFTWARE")
  IMAGE_META_DESCRIPTION = os.getenv("IMAGE_META_DESCRIPTION")
  try:
    logger.info(f"upload_remote_image: {url}")
    content = requests.get(url, params=params).content
    # 视频或音频不压缩
    if file_extension == 'mp4' or file_extension == 'webm' or file_extension == 'mp3':
      logger.info(f"upload_remote_image: {path} is video, no compression")
      return path, upload_file(content, path)
    image = Image.open(BytesIO(content))
    logger.info(f"image.mode: {image.mode}")
    # 动图不压缩
    if is_animated_webp(image):
      logger.info(f"upload_remote_image: {path} is animated webp, no compression")
      return path, upload_file(content, path)
    else:
      logger.info(f"upload_remote_image: {path} is image, compressing")
      dir_path = "output/" + "/".join(path.split("/")[:-1])
      create_dir_if_not_exists(dir_path)
      temp_file_path = "output/" + path
      if image.mode == 'RGBA':
        logger.info(f"upload_remote_image: {path} is RGBA, converting to webp")
        try:
          image.save(temp_file_path, format='PNG', lossless=True, transparency=image.info.get('transparency'))
        except Exception as e:
          logger.error(f'upload_remote_image failed: {e}')
          image.save(temp_file_path, format='PNG')
      else:
        image.save(temp_file_path, format='PNG')
      # image.save(temp_file_path)
      temp_webp_path = temp_file_path.split(".")[0] + ".webp"
      if compress:
        compress_png(temp_file_path, temp_webp_path)
      else:
        temp_webp_path = temp_file_path
      now = datetime.now().strftime("%Y:%m:%d %H:%M:%S")
      add_img_source(temp_webp_path, temp_webp_path, {
        "description": IMAGE_META_DESCRIPTION,
        "artist": IMAGE_META_ARTIST,
        "software": IMAGE_META_SOFTWARE,
        "copyright": IMAGE_META_COPYRIGHT,
        "DateTimeOriginal": now,
        "DateTime": now
      })
      new_content = open(temp_webp_path, "rb").read()
      if compress:
        new_path = path.split(".")[0] + ".webp"
      else:
        new_path = path
    return new_path, upload_file(new_content, new_path)
  except Exception as e:
    logger.error(f'upload_remote_image failed: {e}')
    return path, upload_file(content, path)
  
def debounce(wait_time):
    def decorator(func):
        def debounced_function(*args, **kwargs):
            def call_it():
                debounced_function.timer = None
                return func(*args, **kwargs)
            
            if debounced_function.timer is not None:
                debounced_function.timer.cancel()
            debounced_function.timer = Timer(wait_time, call_it)
            debounced_function.timer.start()
        
        debounced_function.timer = None
        return debounced_function
    return decorator
  
def upload_image_to_comfyui(upload_url, filename, image_url):

  # 下载图片
  response = requests.get(image_url)
  image_data = response.content

  dir_path = datetime.now().strftime("%Y-%m-%d")
  file_id = generate('1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', 10)
  filename = f'{dir_path}-{file_id}-{filename}'

  # 准备 multipart/form-data 请求体
  files = {
      'image': (filename, image_data, 'image/png'),
      'subfolder': (None, ''),
      'type': (None, 'input')
  }

  # 发送 POST 请求上传图片
  resp = requests.post(upload_url, files=files)

  # 检查响应
  if resp.ok:
      res = resp.json()
      logger.info(f'Image uploaded successfully: {res}')
      return res
  else:
      raise Exception(f'Failed to upload image: {resp.text}')
      
def safe_json_loads(json_str):
  try:
    return json.loads(json_str)
  except Exception as e:
    logger.error(f'safe_json_loads failed: {e}')
    return None


import pytest
from unittest.mock import patch, MagicMock
from awe.clients.comfyui_client import ComfyUIClient

def test_comfyui_client_init():
    """测试 ComfyUI 客户端初始化"""
    client = ComfyUIClient("localhost", "8188")
    assert client.host == "localhost"
    assert client.port == "8188"
    assert client.status == "INIT"

@patch('websocket.WebSocket')
def test_comfyui_client_connect(mock_websocket):
    """测试 WebSocket 连接"""
    mock_ws = MagicMock()
    mock_websocket.return_value = mock_ws
    
    client = ComfyUIClient("localhost", "8188")
    client.connect()
    
    mock_ws.connect.assert_called_once()
    assert client.ws is not None

def test_comfyui_client_progress():
    """测试进度计算"""
    client = ComfyUIClient("localhost", "8188")
    
    # 测试空节点列表
    client.current_total_node_ids = []
    assert client.get_progress() == 1
    
    # 测试正常进度计算
    client.current_total_node_ids = ["1", "2", "3", "4"]
    client.current_executed_node_ids = ["1", "2"]
    assert client.get_progress() == 50
    
    # 测试进度为0的情况
    client.current_executed_node_ids = []
    assert client.get_progress() == 1  # 应该返回1而不是0 